#!/usr/bin/python3
# -*- coding: utf-8 -*-
import os
import sys
import re
import xml.dom.minidom
import xacro4sdf.xml_format

#global vaiables
g_property_table = {}
l_property_table = {}
g_macro_params_table = {}
g_macro_node_table = {}

def try2number(str):
    try:
        return float(str)
    except ValueError:
        return str

def parse_url(url):
    path = ""
    result=url.split("://")
    if len(result)!=2:
        return path
    #get file path
    if result[0]=="file":
        if os.path.isfile(result[1]):
            path=result[1]
        elif os.path.isfile("/"+result[1]):
            path = "/"+result[1]
    if result[0]=="model":
        model_paths=[]
        if os.getenv("IGN_GAZEBO_RESOURCE_PATH") is not None:
            model_paths=model_paths+os.getenv("IGN_GAZEBO_RESOURCE_PATH").split(":")
        if os.getenv("GAZEBO_MODEL_PATH") is not None:
            model_paths=model_paths+os.getenv("GAZEBO_MODEL_PATH").split(":")
        for path in model_paths:
            file_path = path+"/"+result[1]
            if(os.path.isfile(file_path)):
                path = file_path
                break
    return path

def get_xacro(root):
    # only find in <sdf>...</sdf>
    for node in root.childNodes:
        if node.nodeType == xml.dom.Node.ELEMENT_NODE:
            if node.tagName == 'xacro_define_property':
                name = node.getAttribute("name")
                g_property_table[name] = try2number(node.getAttribute("value"))
                root.removeChild(node)
            elif node.tagName == 'xacro_define_macro':
                name = node.getAttribute("name")
                g_macro_params_table[name] = node.getAttribute("params").split(' ')
                g_macro_node_table[name] = node.toxml()
                root.removeChild(node)

def get_include_xacro(root):
    for node in root.childNodes:
        if node.nodeType == xml.dom.Node.ELEMENT_NODE:
            if node.tagName == 'xacro_include_definition':
                url = node.getAttribute("url")
                path = parse_url(url)
                #get xacro from file
                if path != "":
                    tmp_doc = xml.dom.minidom.parse(path)
                    get_xacro(tmp_doc.documentElement)
                else:
                    print("not find xacro_include_definition url",url)
                root.removeChild(node)

def replace_inlcude_model_node(node):
    url = node.getAttribute("url")
    path = parse_url(url)
    #get xacro from file
    if path != "":
        parent = node.parentNode
        new_doc = xml.dom.minidom.parse(path).documentElement
        new_nodes = new_doc.getElementsByTagName("model")
        new_node=new_nodes[0]
        for cc in list(new_node.childNodes):
            parent.insertBefore(cc, node)
    else:
        print("not find xacro_include_define url",url)
    parent.removeChild(node)


def re_eval_fn(obj):
    result = eval(obj.group(1), g_property_table, l_property_table)
    return str(result)

def eval_text(xml_str):
    pattern = re.compile(r'[$][{](.*?)[}]', re.S)
    return re.sub(pattern, re_eval_fn, xml_str)

def replace_macro_node(node):
    parent = node.parentNode
    if not node.hasAttribute("name"):
        print("check <xacro_macro> block,not find attrname!")
        sys.exit(1)
    name = node.getAttribute("name")
    # get xml string
    xml_str = g_macro_node_table[name]
    # get local table
    l_property_table.clear()
    for param in g_macro_params_table[name]:
        l_property_table[param] = try2number(node.getAttribute(param))
    # replace macro(insert and remove)
    xml_str = eval_text(xml_str)
    new_node = xml.dom.minidom.parseString(xml_str).documentElement
    for cc in list(new_node.childNodes):
        parent.insertBefore(cc, node)
    parent.removeChild(node)

#reference: https://github.com/ros/xacro/blob/noetic-devel/src/xacro/__init__.py
def addbanner(doc,input_file_name):
    # add xacro auto-generated banner
    banner = [xml.dom.minidom.Comment(c) for c in
              [" %s " % ('=' * 83),
               " |    This document was autogenerated by xacro4sdf from %-26s | " % input_file_name,
               " |    EDITING THIS FILE BY HAND IS NOT RECOMMENDED  %-30s | " % "",
               " %s " % ('=' * 83)]]
    first = doc.firstChild
    for comment in banner:
        doc.insertBefore(comment, first)

def xacro4sdf(inputfile, outputfile):
    doc = xml.dom.minidom.parse(inputfile)
    root = doc.documentElement
    ###### get xacro (common xacro,inlcue xacro,the doc xacro)
    # get common xacro
    folder = os.path.dirname(os.path.abspath(__file__)) 
    get_xacro(xml.dom.minidom.parse(os.path.join(folder,'common.xacro')).documentElement)
    # get inlcude xacro
    get_include_xacro(root)
    # get xacro
    get_xacro(root)
    ####### relapce xacro
    # replace xacro include model
    for _ in range(5):
        nodes = doc.getElementsByTagName("xacro_include_model")
        if nodes.length == 0:
            break
        else:
            for node in list(nodes):
                replace_inlcude_model_node(node)
    if doc.getElementsByTagName("xacro_inlcude_model").length != 0:
        print("Error:The nesting of xacro_include_model is much deep! only support <=5")
        sys.exit(1)
    # replace xacro property (doc -> doc, process by string regular expression operations)
    l_property_table.clear()
    xml_str = root.toxml()
    xml_str = eval_text(xml_str)
    doc = xml.dom.minidom.parseString(xml_str)
    # replace xacro macro
    for _ in range(5):
        nodes = doc.getElementsByTagName("xacro_macro")
        if nodes.length == 0:
            break
        else:
            for node in list(nodes):
                replace_macro_node(node)
    #check
    if doc.getElementsByTagName("xacro_macro").length != 0:
        print("Error:The nesting of macro defination is much deep! only support <=5")
        sys.exit(1)
    # output
    addbanner(doc,inputfile)
    try:
        with open(outputfile, 'w', encoding='UTF-8') as f:
            doc.writexml(f, indent='', addindent='\t', newl='\n', encoding='UTF-8')
    except Exception as err:
        print('output error:{0}'.format(err))

def main():
    if(len(sys.argv) >= 2):
        inputfile = sys.argv[1]
        outputfile = os.path.splitext(inputfile)[0]
        if os.path.splitext(inputfile)[1] == '.xacro':
            # run xacro4sdf
            xacro4sdf(inputfile, outputfile)
            return 0
    #error
    print("usage: xacro4sdf <inputfile> (the name of inputfile must be xxx.xacro)")

if __name__ == '__main__':
    main()